// name: Mover.C
// author: J. Michael Word
// date written: 2/7/98
// purpose: utility routines for bonding and geometric relationships

// **************************************************************
// NOTICE: This is free software and the source code is freely
// available. You are free to redistribute or modify under the
// conditions that (1) this notice is not removed or modified
// in any way and (2) any modified versions of the program are
// also available for free.
//               ** Absolutely no Warranty **
// Copyright (C) 1999 J. Michael Word
// **************************************************************

#include "Mover.h"

// ResBlk.h is included to define sameres()
#include "ResBlk.h"

void Mover::resetFlipMaxScore(int f) {
     _flipMaxScore[f!=0] = LowestMoverScore;
   _flipMaxBadBump[f!=0] = FALSE;
}

void Mover::trackFlipMaxScore(int f, double val, bool hasBadBump) {
   if ( _flipMaxScore[f!=0] < val) {
        _flipMaxScore[f!=0] = val;
      _flipMaxBadBump[f!=0] = hasBadBump;
   }
}

Seq<PDBrec> bondedList(const PDBrec& a, Seq<PDBrec> nearby, int nbnds,
	 Seq<PDBrec> atmList) {
   resetMarks(nearby);
   countBonds(a, nearby, 1, nbnds, atmList); // up to nbnds bonds away

   Seq<PDBrec> bondedAtoms;

   while (nearby) {
      if ((*nearby).mark() >= 1 && (*nearby).mark() <= nbnds) {
	 bondedAtoms.insert(*nearby);
      }
      nearby++;
   }
   return bondedAtoms;
}

void countBonds(const PDBrec& src,
	       const Seq<PDBrec>& nearby, int distcount, int maxcnt,
	       Seq<PDBrec> atmList) {

   Seq<PDBrec> newlyMarked;
   Seq<PDBrec> remainder;

   for (Seq<PDBrec> targ = nearby; targ; targ++) {
      if ( ((*targ).mark() < 1 || (*targ).mark() > distcount)
         && ! diffAltLoc(src, *targ) ) {

	 bool isnear   = withinCovalentDist(src, *targ,  0.2);
	 bool tooclose = withinCovalentDist(src, *targ, -0.5);

	 if (isnear && ! (tooclose || impossibleCovalent(src, *targ, atmList))) {
	    targ.hd().mark(distcount);
	    newlyMarked.insert(*targ);
	 }
	 else {
	    remainder.insert(*targ);
	 }
      }
   }
   if (distcount < maxcnt) {
      for( ; newlyMarked; newlyMarked++) {
	 countBonds((*newlyMarked), remainder, distcount+1, maxcnt, atmList);
      }
   }
}

// initialize the markers we use to determine bonding patterns
void resetMarks(Seq<PDBrec> lst) {
   while (lst) {
      lst.hd().mark(0);
      lst++;
   }
}

bool visableAltConf(const PDBrec& a, bool onlyA) {
   const char aalt = a.alt();
   return (aalt == ' ' || aalt == 'A' || aalt == 'a' || !onlyA);
}

bool interactingConfs(const PDBrec& a, const PDBrec& b, bool onlyA) {
   return ( (!visableAltConf(a, onlyA)) || (!visableAltConf(b, onlyA)) )
      ? FALSE
      : (! diffAltLoc(a, b));
}

// are a and b alternative locations for the same residue?
bool diffAltLoc(const PDBrec& a, const PDBrec& b) {
   return a.resno() == b.resno() && a.insCode() == b.insCode()
       && a.chain() == b.chain()
       && a.alt() != b.alt() && a.alt() != ' ' && b.alt() != ' ';
}

int withinCovalentDist(const PDBrec& p, const PDBrec& q, double offset) {
   double lim = p.covRad() + q.covRad() + offset;

   return distanceSquared(p.loc(), q.loc()) <= (lim*lim);
}

bool impossibleCovalent(const PDBrec& src, const PDBrec& targ, Seq<PDBrec> atmList) {
   if (src.isHydrogen() && targ.isHydrogen()) { return TRUE; }

   else if (src.isHydrogen() || targ.isHydrogen()) {
      bool srcInList = FALSE, targInList = FALSE;
      for( ; atmList; atmList++) {
	      if (*atmList == src)  {srcInList  = TRUE; }
	 else if (*atmList == targ) {targInList = TRUE; }
      }

      if (srcInList && targInList) { return FALSE; }
      else if (srcInList && src.isHydrogen())   { return TRUE; }
      else if (targInList && targ.isHydrogen()) { return TRUE; }
   }

   return (! sameres(src, targ)) && (src.isHydrogen() || targ.isHydrogen());
}

bool foundInList(const PDBrec& a, Seq<PDBrec> lst) {
   while(lst) {
      if (a == *lst) { return TRUE; }
      lst++;
   }
   return FALSE;
}

// what is the gap between the VDW radii of atom a at point aloc
// and atom b at point bloc?

double vdwGap(const PDBrec& a, const Point3d& aloc,
              const PDBrec& b, const Point3d& bloc) {
   double lim = a.vdwRad() + b.vdwRad();

   return distance(aloc, bloc) - lim;
}

// functions used to restrict annular rings of good dots around clashes

bool annularDots(const Point3d& dot, const PDBrec& src, const PDBrec& targ, float probeRadius) {
   return dot2srcCenter(dot, src, targ) >
      kissEdge2bullsEye(src.vdwRad(), targ.vdwRad(), probeRadius);
}

double dot2srcCenter(const Point3d& dot, const PDBrec& src, const PDBrec& targ) {

   const Point3d src2targVec = (targ.loc() - src.loc()).scaleTo(src.vdwRad());
   const Point3d srcSurfacePoint = src2targVec + src.loc();
   return distance(dot, srcSurfacePoint);
}

double kissEdge2bullsEye(float ra, float rb, float rp) {
   return 2.0*ra*sqrt(rb*rp/((ra+rb)*(ra+rp)));
}

// force sorting of Seq<MoverPtr> by descr()
// by duplicating some merge-sort code

Seq<MoverPtr> mpSeqMerge(Seq<MoverPtr> x, Seq<MoverPtr> y) {
   Seq<MoverPtr> r;
   while (x && y) {
      if ( (x.hd()) && (y.hd()) &&
      (x.hd()->descr() < y.hd()->descr())) {
         r.insert(x.hd());
         x++;
      } else {
         r.insert(y.hd());
         y++;
      }
   }
   while (x) {
      r.insert(x.hd());
      x++;
   }
   while (y) {
      r.insert(y.hd());
      y++;
   }
   return r.rev();
}

void mpSeqSplit(Seq<MoverPtr> x, Seq<MoverPtr>& y, Seq<MoverPtr>& z) {
   while(x) {
      y.insert(x.hd());
      if (++x) {
         z.insert(x.hd());
         x++;
      }
   }
}

Seq<MoverPtr> mpSeqSort(const Seq<MoverPtr>& x) {
   if (!x || !x.tl()) { return x; }
   Seq<MoverPtr> p, q;
   mpSeqSplit(x, p, q);
   return mpSeqMerge(mpSeqSort(p), mpSeqSort(q));
}
