// name: RotDonor.C
// author: J. Michael Word
// date written: 2/7/98
// purpose: Implementation for RotDonor

// **************************************************************
// NOTICE: This is free software and the source code is freely
// available. You are free to redistribute or modify under the
// conditions that (1) this notice is not removed or modified
// in any way and (2) any modified versions of the program are
// also available for free.
//               ** Absolutely no Warranty **
// Copyright (C) 1999 J. Michael Word
// **************************************************************

#ifdef OLD_STD_HDRS
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#else
#include <cstdio>
#include <cctype>
#include <cmath>
#endif

#include "RotDonor.h"
#include "AtomPositions.h"

// for isHBDonorOrAcceptorFlipped
#include "FlipMemo.h"

#define START_ANGLE 180.0
#define ROUGH_STEP   10
#define FINE_STEP     1

#define OVERLAPCUTOFF_1    -0.001
#define OVERLAPCUTOFF_3    -0.7
#define OVERLAPCUTOFF_5    -0.9

RotDonor::RotDonor(const Point3d& a, const Point3d& b,
                   const double ang, const PDBrec& heavyAtom)
      : _p1(a), _p2(b), _angle(ang), _heavyAtom(heavyAtom), _nori(0) {
   validateMemo();
}

void RotDonor::finalize(int nBondCutoff, bool useXplorNames,
                        AtomPositions &xyz, DotSphManager& dotBucket) {
   int i = 0;

   const double origAngle = angle();

   if (isComplete()) {

      // pre-build lists of bonded atoms

      const double approxNbondDistLimit = 3.0 + 0.5*nBondCutoff; // just a rule of thumb
      _bnded.resize(_rot.len()+1);
      i = 0;
      Seq<PDBrec> rotAtmList(_heavyAtom, _rot);
      for(Seq<PDBrec> alst = rotAtmList; alst; alst++) {
	 const PDBrec& thisAtom = *alst;
	 _bnded[i++] = bondedList(thisAtom,
	                          xyz.neighbors(thisAtom.loc(),
	                                        thisAtom.covRad(),
	                                        approxNbondDistLimit),
	                          nBondCutoff, rotAtmList);
      }
   // -------------------------------------------------------
   // determine all the angles we will consider

   // first identify nearby acceptors

      Seq<double> angs;

      const ElementInfo& elemHpol = * ElementInfo::StdElemTbl().element("Hpol");

      // O-H & N-H bond len == 1.0, S-H bond len == 1.3
      const double XHbondlen = (_heavyAtom.elem().atno() == 16) ? 1.3 : 1.0;

      for(Seq<PDBrec> nearby = xyz.neighbors(_heavyAtom.loc(), 0.001, 4.0);
	     nearby; nearby++) {

	 if ((*nearby).hasProp(ACCEPTOR_ATOM)
	 || FlipMemo::isHBDonorOrAcceptorFlipped(*nearby, useXplorNames)) {

	    double HBoverlap = distance(_heavyAtom.loc(), (*nearby).loc())
			- ( elemHpol.explRad() + (*nearby).vdwRad()
			   + XHbondlen );

	    if (HBoverlap < OVERLAPCUTOFF_1
	       && ! diffAltLoc(_heavyAtom, *nearby)) {

	       const double ang = origAngle +
		  dihedral((*nearby).loc(), _p1, _p2, (*_rot).loc());

	       angs.insert(clampAngle(ang));
	       if (HBoverlap < OVERLAPCUTOFF_3) {
		  angs.insert(clampAngle(ang - 1.5*ROUGH_STEP));
		  angs.insert(clampAngle(ang + 1.5*ROUGH_STEP));
	       }
	       if (HBoverlap < OVERLAPCUTOFF_5) {
		  angs.insert(clampAngle(ang - 3.0*ROUGH_STEP));
		  angs.insert(clampAngle(ang + 3.0*ROUGH_STEP));
	       }
	    }

	 }
      }

   // -------------------------------------------------------
   // next, look for least-bumping angle not near an acceptor

      double leastBumpAngle = START_ANGLE;
      double bestScore = LowestMoverScore;

      for (i=0; i < (360/ROUGH_STEP); i++) {
	 const double oldTheta = angle();
	 const double theta = START_ANGLE +
	    ((i+1)>>1) * ((i&1) ? 1.0 : -1.0) * ROUGH_STEP;

	 bool tryThisAngle = TRUE;
	 for(Seq<double> chi = angs; chi; chi++) {
	    const double delta = abs(clampAngle(clampAngle(theta) - (*chi)));
	    if (delta < ROUGH_STEP) {
	       tryThisAngle = FALSE;
	       break;
	    }
	 }

	 if (tryThisAngle) {
	    for(Seq<PDBrec> hydlist = _rot; hydlist; hydlist++) {
	       setHydAngle(*hydlist, oldTheta, theta, xyz);
	    }
	    angle(theta);
	    const double bs = bumpsThisAngle(xyz, dotBucket);
	    if (bs > bestScore) { bestScore = bs; leastBumpAngle = theta; }
	 }
      }

      // re-position to least bump angle
      for(Seq<PDBrec> hydlist = _rot; hydlist; hydlist++) {
	 setHydAngle(*hydlist, angle(), leastBumpAngle, xyz);
      }
      angle(leastBumpAngle);
      angs.insert(clampAngle(leastBumpAngle));

   // -------------------------------------------------------
   // merge both sets of angles into the orientation angle array

      _nori = angs.len();
      _oang.resize(_nori);
      int j = 0;
      for(Seq<double> chi = sort(angs).rev(); chi; chi++) {
	 if (j > 0) {
	    const double delta = abs(_oang[j-1] - (*chi));
	    if ((delta >  (0.5*ROUGH_STEP))
	     && (*chi  >=((0.5*ROUGH_STEP)-180.0)) ) {
	       _oang[j++] = *chi;
	    }
	 }
	 else { _oang[j++] = *chi; }
      }
      _nori = j; // <-- the actual length after the merge operation
      _oang.resize(_nori);
   }
}

int RotDonor::makebumpers(MultiDict<LocBlk, BumperPoint>& bblks,
      int rn, float& maxVDWrad) {
   int an = 0;
   const double dtheta = 10.0; // fineness of rotation angle scan
   
   const double scanAngle = 180.0;
   for (Seq<PDBrec> a = _rot; a; a++) {
      for (double theta = -scanAngle; theta < scanAngle; theta += dtheta) {
	 Point3d p((*a).loc().rotate(theta, _p2, _p1));
	 bblks.put(LocBlk(p), BumperPoint(p, rn, an++, (*a).vdwRad()));
	 if ((*a).vdwRad() > maxVDWrad) { maxVDWrad = (*a).vdwRad(); }
      }
   }
   return an;
}

int RotDonor::numOrientations(SearchStrategy ss) const {
   return (ss==Mover::LOW_RES) ?
      _nori :
      (int)(2 * (double(ROUGH_STEP)/double(FINE_STEP)) + 0.5);
}

bool RotDonor::setOrientation(int oi, float delta, AtomPositions &xyz,
                                                       SearchStrategy ss) {
   const double oldTheta = angle();
   const double    theta = orientationAngle(oi, ss) + delta;

   if (abs(theta-oldTheta) > 0.1) {
      for(Seq<PDBrec> hydlist = _rot; hydlist; hydlist++) {
	 setHydAngle(*hydlist, oldTheta, theta, xyz);
      }
      angle(theta);
   }
   rememberOrientation(oi, ss);
   return TRUE;
}

double RotDonor::orientationAngle(int oi, SearchStrategy ss) const {
   double delta = 0.0;
   if (ss != Mover::LOW_RES) {
      const int oh = oi;
      oi = orientation(Mover::LOW_RES);
      delta = ((oh+1)>>1) * ((oh&1) ? 1.0 : -1.0) * FINE_STEP;
   }

   const double a = ((oi >= 0 && oi < _nori) ? _oang[oi] : START_ANGLE)
      + delta;

   return clampAngle(a);
}

Stringclass RotDonor::describeOrientation() const {
   char descrbuf[25];
   ::sprintf(descrbuf, "   rot %4.0f", _angle);
   return Stringclass(descrbuf);
}

double RotDonor::determineScore(AtomPositions &xyz,
	DotSphManager& dotBucket, int nBondCutoff, float probeRadius,
	float pmag, double& penalty, float &bumpScore, float &hbScore, bool &hasBadBump) {

   bumpScore  = 0.0;
   hbScore    = 0.0;
   hasBadBump = FALSE;

   double bestScore = scoreThisAngle(xyz, dotBucket,
                         nBondCutoff, probeRadius,
                         bumpScore, hbScore,
			 hasBadBump);

   penalty = orientationPenalty(pmag);
   return bestScore;
}

double RotDonor::scoreThisAngle(AtomPositions &xyz,
	DotSphManager& dotBucket, int, float probeRadius,
	float &bumpScore, float &hbScore, bool &hasBadBump) {
   const double maxVDWrad = ElementInfo::StdElemTbl().maxExplicitRadius();

   bumpScore  = 0.0;
   hbScore    = 0.0;
   hasBadBump = FALSE;

   double scoreThisO = 0.0;
   int i = 0;
   for(Seq<PDBrec> alst(_heavyAtom, _rot); alst; alst++) {
      float bumpSubScore = 0.0;
      float hbSubScore   = 0.0;
      bool subBadBump    = FALSE;
      const PDBrec& thisAtom = *alst;
      Seq<PDBrec> nearby = xyz.neighbors(thisAtom.loc(), 0.001,
				 thisAtom.vdwRad() + probeRadius + maxVDWrad);

      double val = xyz.atomScore(thisAtom, thisAtom.loc(), nearby,
		_bnded[i], dotBucket.fetch(thisAtom.vdwRad()), probeRadius, FALSE,
		bumpSubScore, hbSubScore, subBadBump);

      bumpScore  += bumpSubScore;
      hbScore    += hbSubScore;
      if (subBadBump) { hasBadBump = TRUE; }

#ifdef DEBUGSUBSCORE
   cerr << "\t:" << PDBrecNAMEout(thisAtom)
      <<":"<< describeOrientation()
      << ": bump=" << bumpSubScore
      << ", HB=" << hbSubScore
      << ((subBadBump==TRUE)?", BADBUMP":"")
//      << ", tot=" << val
      << "\t"
//      << thisAtom.loc()
      << endl;
#endif
      scoreThisO += val;
      i++;
   }

   initializeScoreIfNotSet(scoreThisO, hasBadBump);
   return scoreThisO;
}

double RotDonor::bumpsThisAngle(AtomPositions &xyz, DotSphManager& dotBucket) {
   const double maxVDWrad = ElementInfo::StdElemTbl().maxExplicitRadius();

   double bumpsThisA = 0.0;
   int i = 0;
   for(Seq<PDBrec> alst(_heavyAtom, _rot); alst; alst++) {
      float bumpSubScore = 0.0;
      float hbSubScore   = 0.0;
      bool subBadBump    = FALSE;
      const PDBrec& thisAtom = *alst;
      Seq<PDBrec> nearby = xyz.neighbors(thisAtom.loc(), 0.001,
				 thisAtom.vdwRad() + maxVDWrad);

      double val = xyz.atomScore(thisAtom, thisAtom.loc(), nearby,
		_bnded[i], dotBucket.fetch(thisAtom.vdwRad()), 0.0, TRUE,
		bumpSubScore, hbSubScore, subBadBump);
#ifdef DEBUGSUBSCORE
      cerr <<"\tminimize clashes :"
      << PDBrecNAMEout(thisAtom)
      <<":"<< describeOrientation()
      <<": bump="<<bumpSubScore << ((subBadBump==TRUE)?", BADBUMP":"") <<"\t"
//      << thisAtom.loc()
      << endl;
#endif
      bumpsThisA += val;
      i++;
   }

   return bumpsThisA;
}

void RotDonor::setHydAngle(double newAng, AtomPositions &xyz) {
   const double oldAng = angle();
   for(Seq<PDBrec> hydlist = _rot; hydlist; hydlist++) {
      setHydAngle(*hydlist, oldAng, newAng, xyz);
//cerr  << "RotDonor::setHydAngle[" << (*hydlist).atomname() << "] "
//      << oldAng << " deg --> "
//      << newAng << " deg" << endl;
   }
   angle(newAng);
}

void RotDonor::setHydAngle(PDBrec theAtom, double oldAng, double newAng,
                        AtomPositions &xyz) {
   if (abs(newAng-oldAng) > 0.0001) {
      Point3d lastloc = theAtom.loc();
      theAtom.loc( lastloc.rotate(newAng-oldAng, _p2, _p1) );
      xyz.reposition(lastloc, theAtom);
   }
}
